
#include "prefix.sh"
#include "texture_utility.sh"
#include "dof.sh"

uniform highp sampler2D s_cc3Texture2Ds[3];
uniform vec4 u_samplingPattern[24];

varying vec2 v_texCoord;

vec3 sampleHighRes(vec2 offset) {
    return sampleLevelZero(s_cc3Texture2Ds[0], v_texCoord + offset).rgb;
}

float sampleDepthHighRes(vec2 offset) {
    return sampleLevelZero(s_cc3Texture2Ds[1], v_texCoord + offset).x;
}

vec4 sampleLowRes(vec2 offset) {
    return sampleLevelZero(s_cc3Texture2Ds[2], v_texCoord + offset);
}

vec2 getSampleOffset(int index, bool xJitter, bool yJitter) {
    int sampleOffset = xJitter ? 0 : 12;
    vec4 v4 = u_samplingPattern[sampleOffset + index];
    return yJitter ? v4.xy : v4.zw;
}

void main() {
    vec4 centerTap = vec4(
        sampleHighRes(vec2(0.0, 0.0)),
        depthValueToCircleOfConfusion(sampleDepthHighRes(vec2(0.0, 0.0))));
    float centerCoc = centerTap.a;
    float cocRadius = centerCoc * u_dofParams.z;

    // Use a 2x2 pattern for jittering the sample pattern. This introduces a little
    // grainness, but can help reduce the number of taps required per sample.
    bool xJitter = fract(gl_FragCoord.x * 0.5) < 0.5;
    bool yJitter = fract(gl_FragCoord.y * 0.5) < 0.5;

    vec4 result = vec4(centerTap.rgb * centerTap.a, centerTap.a);
    for (int c=0; c<12; ++c) {
        vec2 stochasticOffset = getSampleOffset(c, xJitter, yJitter) * cocRadius;
        vec4 tapHigh = vec4(
            sampleHighRes(stochasticOffset).rgb,
            depthValueToCircleOfConfusion(sampleDepthHighRes(stochasticOffset)));
        vec4 tapLow = sampleLowRes(stochasticOffset * 0.25);    // (assuming 16:1 downsample for low res image)

        // as per Scheuermann04, mix low and high res taps
        vec4 mixedTap = mix(tapHigh, tapLow, tapHigh.a);

        mixedTap.a = (mixedTap.a >= centerTap.a) ? 1.0 : mixedTap.a;

        result.rgb += mixedTap.rgb * mixedTap.a;
        result.a += mixedTap.a;
    }

    gl_FragColor.rgb = result.rgb / result.a;
    gl_FragColor.a = centerTap.a;    // write out an approximation of the effect that depth of field has had on this pixel
}


