//#define DRAW_IN_GRAYSCALE
//#define OVERDRAW
//#define ALPHA_TEST
//#define DEBUG

#include "prefix.sh"

#include "programmable_blending.sh"

VARYING_IN vec2 v_texCoord;
VARYING_IN vec4 v_color;
VARYING_IN vec2 v_time;

uniform sampler2D un_diffuseMap;
uniform vec4 un_fragPacked[4];

vec4  getDiffuseColor()         { return un_fragPacked[0]; }
vec4  getClampValues()          { return un_fragPacked[1]; }
float getOpacity()              { return un_fragPacked[2].x; }
bool  getUseColor()             { return bool(un_fragPacked[2].y > 0.5); }
bool  getMultiplyColorWithRGB() { return bool(un_fragPacked[2].z > 0.5); }
float getAlphaTestValue()       { return un_fragPacked[2].w; }

#if !defined(GL_ES)
uniform float un_hideShadedWireframe;
float getHideShadedWireframe()  { return un_hideShadedWireframe; }
#endif

void main() {
    vec4 texel = texture2D(un_diffuseMap, v_texCoord);
    vec3 colorMultiplier = mix(vec3(1.0), texel.rgb, float(getMultiplyColorWithRGB()));
    gl_FragColor = vec4(mix(texel.rgb, v_color.rgb * colorMultiplier, float(getUseColor())),
                        texel.a * v_color.a * getOpacity()) * getDiffuseColor();

    // Apply programmable blending
    apply_programmable_blending();

#if 0 // visualize back faces
    if (!gl_FrontFacing) {
        gl_FragColor.rgb = vec3(1) - gl_FragColor.rgb;
    }
#endif

#if defined(DRAW_IN_GRAYSCALE)
    float color = dot(gl_FragColor.rgb, vec3(0.2126, 0.7152, 0.0722));
    gl_FragColor = vec4(color, color, color, gl_FragColor.a);
#endif

#if defined(OVERDRAW)
    gl_FragColor = vec4(0.25, 0.05, 0.025, 0.08);
#endif

#if !defined(GL_ES)
    // wireframe is enabled with a getHideShadedWireframe()-value of 0, resulting in black lines with full alpha
    gl_FragColor = vec4(gl_FragColor.rgb * getHideShadedWireframe(), gl_FragColor.a + 1.0-getHideShadedWireframe());
#endif

#if defined(DEBUG) || defined(OVERDRAW)
    // If the vertex shader suggested to discard the particle, we discard the fragment
    if (v_color.a == 0.0) {
//        gl_FragColor = vec4(0.0, 1.0, 0.0, 1.0); /* uncomment and do not discard for debug visualization */
        discard;
    }
#endif
#if defined(OVERDRAW)
    return;
#endif

#if defined(ALPHA_TEST)
    if(gl_FragColor.a < getAlphaTestValue()) {
        discard;
    }
#endif
    float clampValue = step(0.5, v_time.y) * mix(getClampValues().y, getClampValues().z, 2.0 * (v_time.y - 0.5)) +
                       (1.0 - step(0.5, v_time.y)) * mix(getClampValues().x, getClampValues().y, 2.0 * v_time.y);
    clampValue *= getClampValues().w;
    if (gl_FragColor.a < clampValue) {
        discard;
    }
}
