//#define BILLBOARD
//#define DIRECTIONAL_BILLBOARD
//#define ROTATE_TO_FACE
//#define AXIAL_BILLBOARDING

// OpenGL ES 3 preamble
#if __VERSION__ >= 300
    #define ATTRIBUTE in
    #define VARYING_IN in
    #define VARYING_OUT out
#else
    #define ATTRIBUTE attribute
    #define VARYING_IN varying
    #define VARYING_OUT varying
#endif

// instancing preamble
#if CC3_PLATFORM_WINDOWS
    // windows
    #define INSTANCE_ID 0
#elif defined(GL_ES)
    // ios + android
    #if __VERSION__ >= 300
        #define INSTANCE_ID gl_InstanceID
    #else
        #extension GL_EXT_draw_instanced: enable
        #define INSTANCE_ID gl_InstanceIDEXT
    #endif
    #define INSTANCING_ENABLED
#else
    // osx
    #extension GL_ARB_draw_instanced: enable
    #define INSTANCE_ID gl_InstanceIDARB
#endif

precision highp float;

ATTRIBUTE vec4 in_position;
ATTRIBUTE vec4 in_speed;
ATTRIBUTE vec4 in_uv_life;
ATTRIBUTE vec4 in_color1;
ATTRIBUTE vec4 in_color2;
ATTRIBUTE vec4 in_color3;
ATTRIBUTE vec4 in_scale;
ATTRIBUTE vec4 in_moreScale;
ATTRIBUTE vec4 in_offsetPosition;
ATTRIBUTE vec4 in_rotation;
ATTRIBUTE vec4 in_modelMatrix_4e0;
ATTRIBUTE vec4 in_modelMatrix_4e1;
ATTRIBUTE vec4 in_modelMatrix_4e2;
ATTRIBUTE vec4 in_textureRandomStart;
ATTRIBUTE vec4 in_positionCoord;

#define MAX_PARTICLE_INSTANCES 25

#ifdef INSTANCING_ENABLED
uniform mat4 un_modelMatrices[MAX_PARTICLE_INSTANCES];
uniform mat4 un_orientationMatrices[MAX_PARTICLE_INSTANCES];
uniform float un_times[MAX_PARTICLE_INSTANCES];
#else
uniform mat4 un_modelMatrix;
uniform mat4 un_orientationMatrix;
uniform float un_time;
#endif

uniform vec4 un_vertPacked[21];

mat4  getViewMatrix()           { return mat4(un_vertPacked[0], un_vertPacked[1], un_vertPacked[2], un_vertPacked[3]); }
mat4  getProjectionMatrix()     { return mat4(un_vertPacked[4], un_vertPacked[5], un_vertPacked[6], un_vertPacked[7]); }
vec3  getNonBillboardOrientationInitialRotation()   { return un_vertPacked[8].xyz; }
vec3  getNonBillboardOrientationInitialSpeed()      { return un_vertPacked[9].xyz; }
vec3  getNonBillboardOrientationFinalSpeed()        { return un_vertPacked[10].xyz; }
float getEmitterDuration()      { return un_vertPacked[8].w; }
float getXZAngularSpeed()       { return un_vertPacked[9].w; }
float getXZAngularAcceleration()    { return un_vertPacked[10].w; }
ivec2 getNumberExtraRowsColumns()   { return ivec2(un_vertPacked[11].w, un_vertPacked[12].w); }
vec3  getRadialSpeed()          { return un_vertPacked[11].xyz; }
vec3  getRadialAcceleration()   { return un_vertPacked[12].xyz; }
vec3  getLinearAcceleration()   { return un_vertPacked[13].xyz; }
vec3  getBillboardAxis()        { return un_vertPacked[14].xyz; }
vec3  getCameraPosition()       { return un_vertPacked[15].xyz; }
vec3  getAnchor()               { return vec3(un_vertPacked[13].w, un_vertPacked[14].w, un_vertPacked[19].w); }
float getTextureRotationAngle() { return un_vertPacked[15].w; }
float getTextureAnimSpeed()     { return un_vertPacked[16].w; }
vec3  getColorTimings()         { return un_vertPacked[16].xyz; }
vec4  getScaleTimings()         { return un_vertPacked[17]; }
vec4  getUVOffset()             { return un_vertPacked[18]; }
float getStretchFactor()        { return un_vertPacked[19].x; }
bool  getStretchYOnly()         { return bool(un_vertPacked[19].y > 0.5); }
float getOvermultiplyAlpha()    { return un_vertPacked[19].z; }
vec2  getFadeRange()            { return un_vertPacked[20].xy; }

VARYING_OUT vec2 v_texCoord;
VARYING_OUT vec4 v_color;
VARYING_OUT vec2 v_time;

uniform float un_erdFaceAxis;
uniform float un_erdStartAngle;
uniform float un_erdAnglePerInstance;
uniform float un_erdSpinRate;

#define M_PI 3.1415926535897932384626433832795
#define M_DEG_TO_RAD    0.0174532925199432957692369076849 // PI / 180

#if (defined(NO_TRACK) && (defined(BILLBOARD) || defined(DIRECTIONAL_BILLBOARD))) || (defined(ROTATE_TO_FACE) && defined(AXIAL_BILLBOARDING))
mat3 matrixTranspose(mat3 matrix) {
    vec3 A = matrix[0];
    vec3 B = matrix[1];
    vec3 C = matrix[2];
    return mat3(vec3(A.x, B.x, C.x),
                vec3(A.y, B.y, C.y),
                vec3(A.z, B.z, C.z));
}

mat4 rotationInverse(mat4 matrix) {
    mat3 rotationScale = mat3(matrix);
    float scaleX = length(matrix[0]);
    float scaleY = length(matrix[1]);
    float scaleZ = length(matrix[2]);
    mat3 scaleInverse = mat3(vec3(1.0/scaleX, 0, 0),
                             vec3(0, 1.0/scaleY, 0),
                             vec3(0, 0, 1.0/scaleZ));
    // Shader compilation fails on Nexus 6 if following is inlined into return statement
    mat3 matrixInverse = rotationScale * scaleInverse;
    return mat4(matrixTranspose(matrixInverse));
}
#endif

float matrixDeterminant(mat2 m) {
    return m[0][0] * m[1][1] - m[0][1] * m[1][0];
}

float matrixDeterminant(mat3 m) {
    return m[0][0] * matrixDeterminant(mat2(m[1][1], m[1][2], m[2][1], m[2][2])) -
           m[1][0] * matrixDeterminant(mat2(m[0][1], m[0][2], m[2][1], m[2][2])) +
           m[2][0] * matrixDeterminant(mat2(m[0][1], m[0][2], m[1][1], m[1][2]));
}

mat4 expensiveInverse(mat4 m) {
    // ported from CC3Matrix4x4InvertAdjoint
    mat4 adj = mat4(+matrixDeterminant(mat3(m[1].yzw, m[2].yzw, m[3].yzw)),
                    -matrixDeterminant(mat3(m[0].yzw, m[2].yzw, m[3].yzw)),
                    +matrixDeterminant(mat3(m[0].yzw, m[1].yzw, m[3].yzw)),
                    -matrixDeterminant(mat3(m[0].yzw, m[1].yzw, m[2].yzw)),

                    -matrixDeterminant(mat3(m[1].xzw, m[2].xzw, m[3].xzw)),
                    +matrixDeterminant(mat3(m[0].xzw, m[2].xzw, m[3].xzw)),
                    -matrixDeterminant(mat3(m[0].xzw, m[1].xzw, m[3].xzw)),
                    +matrixDeterminant(mat3(m[0].xzw, m[1].xzw, m[2].xzw)),

                    +matrixDeterminant(mat3(m[1].xyw, m[2].xyw, m[3].xyw)),
                    -matrixDeterminant(mat3(m[0].xyw, m[2].xyw, m[3].xyw)),
                    +matrixDeterminant(mat3(m[0].xyw, m[1].xyw, m[3].xyw)),
                    -matrixDeterminant(mat3(m[0].xyw, m[1].xyw, m[2].xyw)),

                    -matrixDeterminant(mat3(m[1].xyz, m[2].xyz, m[3].xyz)),
                    +matrixDeterminant(mat3(m[0].xyz, m[2].xyz, m[3].xyz)),
                    -matrixDeterminant(mat3(m[0].xyz, m[1].xyz, m[3].xyz)),
                    +matrixDeterminant(mat3(m[0].xyz, m[1].xyz, m[2].xyz)));
    float det = (adj[0][0] * m[0][0]) + (adj[0][1] * m[1][0]) + (adj[0][2] * m[2][0]) + (adj[0][3] * m[3][0]);
    return adj / det;
}

#if defined(DIRECTIONAL_BILLBOARD) || defined(ROTATE_TO_FACE)
vec4 quaternionFromAxisAngle(vec3 axis, float angle) {
    if (length(axis) < 1e-4) axis = vec3(0, 1, 0);
    axis = normalize(axis);
    float halfAngle = -angle / 2.0;
    float sinHalfAngle = sin(halfAngle);
    float cosHalfAngle = cos(halfAngle);
    return vec4(axis * sinHalfAngle, cosHalfAngle);
}

vec4 quaternionMultiply(vec4 a, vec4 b) {
    return vec4(a.w * b.x + a.x * b.w + a.y * b.z - a.z * b.y,
                a.w * b.y + a.y * b.w + a.z * b.x - a.x * b.z,
                a.w * b.z + a.z * b.w + a.x * b.y - a.y * b.x,
                a.w * b.w - a.x * b.x - a.y * b.y - a.z * b.z);
}

mat4 mat4FromQuaternion(vec4 q) {
    float twoXX = 2.0 * q.x * q.x;
    float twoXY = 2.0 * q.x * q.y;
    float twoXZ = 2.0 * q.x * q.z;
    float twoXW = 2.0 * q.x * q.w;

    float twoYY = 2.0 * q.y * q.y;
    float twoYZ = 2.0 * q.y * q.z;
    float twoYW = 2.0 * q.y * q.w;

    float twoZZ = 2.0 * q.z * q.z;
    float twoZW = 2.0 * q.z * q.w;

    vec4 col1 = vec4(1.0 - twoYY - twoZZ,
                     twoXY - twoZW,
                     twoXZ + twoYW,
                     0.0);

    vec4 col2 = vec4(twoXY + twoZW,
                     1.0 - twoXX - twoZZ,
                     twoYZ - twoXW,
                     0.0);

    vec4 col3 = vec4(twoXZ - twoYW,
                     twoYZ + twoXW,
                     1.0 - twoXX - twoYY,
                     0.0);
    vec4 col4 = vec4(0.0, 0.0, 0.0, 1.0);
    return mat4(col1, col2, col3, col4);
}
#endif // DIRECTIONAL_BILLBOARD || ROTATE_TO_FACE

#if !defined(BILLBOARD)
mat4 matrixFromRotationYXZ(vec3 rotationInDegrees) {
    // Same as CC3Matrix4x4PopulateFromRotationYXZ
    /**
     * Populates the specified matrix as a rotation around three axes, y (yaw), x (pitch) and z (roll),
     * in that order, from the specified Euler angle rotation. Each Euler angle is specified in degrees.
     *
     * This rotation places 'up' along the positive Y axis, which is the OpenGL ES default.
     */
    /*
     *     |  cycz + sxsysz   czsxsy - cysz   cxsy  |
     * M = |  cxsz            cxcz           -sx    |
     *     |  cysxsz - czsy   cyczsx + sysz   cxcy  |
     *
     *     where cA = cos(A), sA = sin(A) for A = x,y,z
     */
    vec3 rotRads = rotationInDegrees * M_DEG_TO_RAD;
    mat4 m = mat4(1.0); // Begin with identity matrix

    float cx = cos(rotRads.x);
    float sx = sin(rotRads.x);
    float cy = cos(rotRads.y);
    float sy = sin(rotRads.y);
    float cz = cos(rotRads.z);
    float sz = sin(rotRads.z);

    m[0][0] = (cy * cz) + (sx * sy * sz);
    m[0][1] = cx * sz;
    m[0][2] = (cy * sx * sz) - (cz * sy);

    m[1][0] = (cz * sx * sy) - (cy * sz);
    m[1][1] = cx * cz;
    m[1][2] = (cy * cz * sx) + (sy * sz);

    m[2][0] = cx * sy;
    m[2][1] = -sx;
    m[2][2] = cx * cy;

    m[3][3] = 1.0;

    return m;
}
#endif // #if !BILLBOARD

float atan2(float y, float x) {
    return x == 0.0 ? sign(y) * M_PI * 0.5 : atan(y, x);
}

mat4 currentModelMatrix() {
#ifdef INSTANCING_ENABLED
    return un_modelMatrices[INSTANCE_ID];
#else
    return un_modelMatrix;
#endif
}

mat4 currentOrientationMatrix() {
#ifdef INSTANCING_ENABLED
    return un_orientationMatrices[INSTANCE_ID];
#else
    return un_orientationMatrix;
#endif
}

float currentTime() {
#ifdef INSTANCING_ENABLED
    return un_times[INSTANCE_ID];
#else
    return un_time;
#endif
}

void main() {
    float delay = in_uv_life.w;
    float lifetime = in_uv_life.z;
    float realTime = currentTime() - delay;
    float t = realTime - float(int(realTime / lifetime)) * lifetime;
    float t2 = 0.5 * t * t; // just to save some instructions

    vec3 direction = in_speed.xyz + getLinearAcceleration().xyz * t; // direction of motion of particle

    /* begin scale */
    // there are four keyframes, ordered x,y,z,w (initial, intermediate, second intermediate, final)
    float isScaleBeforeSecond = float(t < getScaleTimings().y * lifetime);
    float isScaleAfterSecondBeforeThird = float(t >= getScaleTimings().y * lifetime) *
                                          float(t < getScaleTimings().z * lifetime);
    float isScaleAfterThird = float(t >= getScaleTimings().z * lifetime);

    float scaleBeforeSecondLerp = (max(t / lifetime, getScaleTimings().x) - getScaleTimings().x) /
                                  (getScaleTimings().y - getScaleTimings().x);
    float scaleAfterSecondBeforeThirdLerp = ((t / lifetime) - getScaleTimings().y) /
                                            (getScaleTimings().z - getScaleTimings().y);
    float scaleAfterThirdLerp = ((min(t / lifetime, getScaleTimings().w) - getScaleTimings().z) /
                                (getScaleTimings().w - getScaleTimings().z));

    vec3 initialScale = in_textureRandomStart.yzw;
    vec3 intermediateScale = vec3(in_scale.xy, in_rotation.x);
    vec3 secondIntermediateScale = in_moreScale.xyz;
    vec3 finalScale = vec3(in_scale.zw, in_rotation.y);
    float particleSize = in_offsetPosition.w;
    // interpolate
    vec3 vectorScale = isScaleBeforeSecond * mix(initialScale, intermediateScale, scaleBeforeSecondLerp) +
                       isScaleAfterSecondBeforeThird * mix(intermediateScale, secondIntermediateScale, scaleAfterSecondBeforeThirdLerp) +
                       isScaleAfterThird * mix(secondIntermediateScale, finalScale, scaleAfterThirdLerp);

    vectorScale *= particleSize;

    vectorScale += getStretchFactor() * abs(direction) * vec3(1.0 - float(getStretchYOnly()), 1.0, 1.0 - float(getStretchYOnly()));
    /* end scale */

    vec3 anchorOffset = 2.0 * vectorScale * (getAnchor() - vec3(0.5, 0.5, 0.5));

    // vertexOffset is xyz of relative coords in (-1, 1) space of this vertex.
    vec3 vertexOffset = in_positionCoord.xyz * vectorScale.xyz - anchorOffset;

    // rotation about origin for polar coordinates
    float xzAngle = getXZAngularSpeed() * t + t2 * getXZAngularAcceleration();

    mat2 xzRotation = mat2(vec2( cos(xzAngle), -sin(xzAngle) ),
                           vec2( sin(xzAngle),  cos(xzAngle) )); // using transpose of rotation matrix because we are rotating x to -z instead of to +z

    vec3 translation = in_position.xyz;
    translation.xz = xzRotation * translation.xz;

    // vec multiplication is pointwise, so this scales about the origin (emitter's coordinates is handled by the cc3node.location)
    translation += normalize(translation) * (t * getRadialSpeed() + t2 * getRadialAcceleration());

    // xyz added afterwards is equivalent to the above polar world being 'windswept'
    translation += in_speed.xyz * t + getLinearAcceleration() * t2;

    translation += in_offsetPosition.xyz;

    // ######## Rotation matrix ########
    /* by our design, local rotation of the sprite in its own plane is mutually exclusive with rotating to face direction of movement (the design could be adjusted if desired) */
    #if !defined(ROTATE_TO_FACE)
        // Local rotation of the sprite in its own plane of x and y
        float rotationSpeed = in_speed.w;
        float initialRotation = in_position.w;
        float angle = initialRotation + rotationSpeed * t;

        mat4 rotationMatrix4D = mat4( cos(angle), sin(angle), 0, 0,
                                     -sin(angle), cos(angle), 0, 0,
                                     0, 0, 1, 0,
                                     0, 0, 0, 1);  // matrix is initialized by columns, so this is a standard counterclockwise rotation matrix around Z
    #else
        vec3 normalizedDirection = normalize(direction);
        /* Math taken from answer to question: How do I generate a rotation matrix to map one vector onto another? */
        // Create rotation matrix that will map 'up' direction of particle (0,1,0) to the direction of movement (direction)
        vec3 axisOfRotation = cross(vec3(0,1,0), normalizedDirection); // axis of rotation between two vectors that has shortest angular distance is the cross product
        float angleOfRotation = acos(dot(vec3(0,1,0), normalizedDirection));
        vec4 q = quaternionFromAxisAngle(axisOfRotation, angleOfRotation);
        mat4 rotationMatrix4D = mat4FromQuaternion(q);
    #endif
    // ######## End rotation matrix ########

    #if !defined(BILLBOARD)
        /* begin non-billboard rotational change */
        // Calculate non-billboard rotation due to initial rotation, speed and acceleration
        vec3 nonBillboardOrientationAcceleration = (getNonBillboardOrientationFinalSpeed() - getNonBillboardOrientationInitialSpeed()) / lifetime; // change in speed over lifespan
        vec3 changeInNonBillboardOrientation = getNonBillboardOrientationInitialRotation() +
                                               (getNonBillboardOrientationInitialSpeed() * t) +
                                               (nonBillboardOrientationAcceleration * t2); // distance = initial_distance + initial_velocity * time + 0.5 * acceleration * time^2

        // Populate matrix with rotation
        mat4 orientationChangeMatrix = matrixFromRotationYXZ(changeInNonBillboardOrientation);
        // Below, we combine current orientation with this orientation change due to speed and acceleration
        /* end non-billboard rotational change */
    #endif

    #if defined(NO_TRACK)
        mat4 modelMatrix = mat4(in_modelMatrix_4e0.xyz, 0.0,
                                in_modelMatrix_4e0.w, in_modelMatrix_4e1.xy, 0.0,
                                in_modelMatrix_4e1.zw, in_modelMatrix_4e2.x, 0.0,
                                in_modelMatrix_4e2.yzw, 1.0);
    #else
        mat4 modelMatrix = currentModelMatrix();
    #endif

    //##########################################################
    //################ Begin orientation matrix ################
    /* Options related to orientation are the following:
     *  - no billboarding
     *  - billboarding toward camera
     *  - directional billboards around x or y, but not both (both were previously supported, but new method is different)
     *  - rotate particle to face the direction of movement
     *    - with billboarding to rotate particle around remaining degree of freedom
     */
    /* Feature requirements
                                                orientation change matrix       uses logic to billboard around an axis
     BILLBOARD                                  NO
     (not billboard)                            YES
     DIRECTIONAL_BILLBOARD                      YES                             YES
     ROTATE_TO_FACE                             YES                             NO
        ROTATE_TO_FACE_ALIGN_CAMERA             YES                             YES
     */
    #if defined(NO_TRACK) && defined(BILLBOARD)
        // calculate billboard matrix because it's different for every quad
        mat4 orientationMatrix = getViewMatrix() * modelMatrix;
        orientationMatrix = rotationInverse(orientationMatrix);
    #elif defined(DIRECTIONAL_BILLBOARD) || (defined(ROTATE_TO_FACE) && defined(AXIAL_BILLBOARDING))
        //##### billboarding around a direction #####
        // The axis for billboarding may be either:
        //  1) the directional billboard axis
        //  2) the direction vector (if rotating to face in the direction of movement and also facing the camera)
        #if defined(DIRECTIONAL_BILLBOARD)
            vec3 axisForBillboarding = getBillboardAxis();
        #elif defined(ROTATE_TO_FACE)
            vec3 axisForBillboarding = normalizedDirection;
        #endif

        /* Note: Most of these comments are intended for rotation to face in the direction of motion.
         * Support for axis-aligned directional billboarding was incorporated after. */

        // We have rotated the vertex in the direction of motion and to billboard around its direction axis,
        // we can work in a simpler space by bringing the camera into the original model space of the vertex.
        // To bring the vertex back into model space, we'd multiply by the inverse of the transform;
        // we can do the same with the camera's position.
        vec4 cameraPositionInWorld = vec4(getCameraPosition(), 1.0); // the view matrix does not have the camera's full translation, but instead has its rotation around the origin and distance from the origin.  Instead of finding the global position with that, we have passed in the global position.
        vec4 cameraPositionInModelSpace = expensiveInverse(modelMatrix) * cameraPositionInWorld - vec4(translation, 0);
        #if defined(ROTATE_TO_FACE)
            mat4 inv = rotationInverse(rotationMatrix4D);
            cameraPositionInModelSpace = inv * cameraPositionInModelSpace;
        #endif
        // We want to rotate the original normal toward the camera.
        // However, that rotation is restricted to the xz plane because the up (y) vector was fixed to point in the direction of motion.
        // We project the camera position onto the xz plane.
        vec4 projectedCameraPositionInModelSpace = cameraPositionInModelSpace; // We don't actually zero out a value; we use two of the three values though, depending on what axis we are billboarding around
        // The rotation angle is the angle between the forward (z) vector and the camera's projected position.
        float angleOfBillboardRotation = atan2(projectedCameraPositionInModelSpace.x, projectedCameraPositionInModelSpace.z); // calculate angle in a zx-coordinate system

        #if defined(DIRECTIONAL_BILLBOARD)
            // For directional billboarding, rotating around y is the same as the above calculation used for billboarding when rotating to face in the direction of movement.
            // However, when rotating around x, the calculation is different.
            // We can 'mix' the two values to choose the correct one.
            /* this assumes that getBillboardAxis() is either (0,1,0) [billboard around y] or (1,0,0) [billboard around x] */
            float angleOfBillboardRotationAroundY = getBillboardAxis().y * angleOfBillboardRotation;
            float angleOfBillboardRotationAroundX = getBillboardAxis().x * -atan2(projectedCameraPositionInModelSpace.y, projectedCameraPositionInModelSpace.z); // calculate angle in a zy-coordinate system
            angleOfBillboardRotation = angleOfBillboardRotationAroundY + angleOfBillboardRotationAroundX;
        #endif

        vec4 billboardQuaternion = quaternionFromAxisAngle(axisForBillboarding, angleOfBillboardRotation);
        mat4 orientationMatrix = mat4FromQuaternion(billboardQuaternion) * orientationChangeMatrix;
        //##### end billboarding around a direction #####
    #elif defined(BILLBOARD)
        // use provided orientation for billboards that track emitter.  (the provided orientation was calculated for billboarding).
        mat4 orientationMatrix = currentOrientationMatrix();
    #else
        // non-billboarding and tracking emitter; non-billboarding and not tracking emitter
        mat4 orientationMatrix = orientationChangeMatrix; // include orientation change due to rotation - initial, speed and acceleration
    #endif
    //################ End orientation matrix ##################
    //##########################################################


    //################ gl_Position ################
    #if defined(USE_EVEN_ROTATION_DISTRIBUTION)
        //Each particle is positioned at an even rotation around the center axis, like so:  *
        //in_positionCoord.xy
        //The "y" vector is the axis of rotation, erd_Axis
        //The "x" axis is any vector which is orthogonal to this vector.  But that's not super easy to calculate.  For now we fake it
        //by just using an arbitrary vector which is highly unlikely to be the vector of our axis
        vec3 yAxis = vec3(0.0, 1.0, 0.0);
        vec3 xAxis = vec3(1.0, 0.0, 0.0);
        float currentAngle = un_erdStartAngle + un_erdSpinRate * currentTime() + in_positionCoord.w * un_erdAnglePerInstance;
        //Now we have to rotate xAxis by currentAngle around yAxis
        //Thank you very much to Neil Mendoza
        float s = sin(currentAngle);
        float c = cos(currentAngle);
        float oc = 1.0 - c;
        mat3 erdRotationMatrix = mat3(c,     0.0,           s,
                                      0.0,   oc + c,        0.0,
                                      -s,    0.0,           c);
        xAxis = normalize(xAxis * erdRotationMatrix);
        vec3 offset = vertexOffset.x * xAxis + vertexOffset.y * yAxis;
        vec4 particlePositionWithoutFace = getProjectionMatrix() * getViewMatrix() * modelMatrix * (vec4(translation, 0) + vec4(offset, 1));

        vec3 faceYAxis = xAxis;
        vec3 faceXAxis = normalize(cross(xAxis, yAxis));
        vec2 faceVertexOffset = (vertexOffset.xy + vectorScale.xy) * 0.5;
        vec3 faceOffset = faceVertexOffset.x * faceXAxis + faceVertexOffset.y * faceYAxis;
        vec4 particlePositionWithFace = getProjectionMatrix() * getViewMatrix() * modelMatrix * (vec4(translation, 0) + vec4(faceOffset, 1));
        gl_Position = un_erdFaceAxis * particlePositionWithFace + (1.0 - un_erdFaceAxis) * particlePositionWithoutFace;
    #else // USE_EVEN_ROTATION_DISTRIBUTION
        gl_Position = getProjectionMatrix() * getViewMatrix() * modelMatrix * (vec4(translation, 0) + orientationMatrix * (rotationMatrix4D * vec4(vertexOffset, 1)));
    #endif
    //################ End gl_Position ################

    // Get texture coordinate based on frame
    int frameIndex = int(float((getNumberExtraRowsColumns().x + 1) * (getNumberExtraRowsColumns().y + 1)) * t * getTextureAnimSpeed() / lifetime) + int(in_textureRandomStart.x);
    int row = frameIndex / (getNumberExtraRowsColumns().y + 1);
    int col = frameIndex - row * (getNumberExtraRowsColumns().y + 1);
    v_texCoord = vec2((float(col) + in_uv_life.x) / float(getNumberExtraRowsColumns().y + 1),
                      (float(getNumberExtraRowsColumns().x - row) + in_uv_life.y) / float(getNumberExtraRowsColumns().x + 1));

    vec2 uvScrollSpeed = in_rotation.zw;

    mat2 textureRotationMatrix = mat2(vec2(cos(getTextureRotationAngle()), -sin(getTextureRotationAngle())),
                                      vec2(sin(getTextureRotationAngle()), cos(getTextureRotationAngle())));
    v_texCoord += uvScrollSpeed * t; // texture scrolling
    v_texCoord += getUVOffset().xy; // uv-offset
    /* rotate around center, not around origin */
    v_texCoord -= vec2(0.5, 0.5); // shift to center
    v_texCoord *= textureRotationMatrix; // texture rotation
    v_texCoord += vec2(0.5, 0.5); // shift to origin
    v_texCoord *= getUVOffset().zw; // uv-tiling

    // Distance fading
    float nearZ_clip = (getProjectionMatrix() * vec4(0, 0, -getFadeRange().x, 1)).z;
    float farZ_clip  = (getProjectionMatrix() * vec4(0, 0, -getFadeRange().y, 1)).z;
    float distanceOpacity = 1.0;
    if (nearZ_clip < farZ_clip) {
        distanceOpacity = smoothstep(0.0, 1.0, (gl_Position.z - nearZ_clip) / (farZ_clip - nearZ_clip));
    }

    // Color interpolation
    float isColorBeforeSecond = float(t < getColorTimings().y * lifetime);
    float isColorAfterSecond = float(t >= getColorTimings().y * lifetime);

    float colorBeforeSecondLerp = (max(t / lifetime, getColorTimings().x) - getColorTimings().x) /
                                  (getColorTimings().y - getColorTimings().x);
    float colorAfterSecondLerp = (min(t / lifetime, getColorTimings().z) - getColorTimings().y) /
                                 (getColorTimings().z - getColorTimings().y);

    v_color = isColorBeforeSecond * mix(in_color1, in_color2, colorBeforeSecondLerp) +
              isColorAfterSecond * mix(in_color2, in_color3, colorAfterSecondLerp);

    // If particle should not be alive at this time, discard it
    float shouldDiscard = float(t < 0.0) + float(getEmitterDuration() > 0.0) * (float(getEmitterDuration() < delay) + float(realTime >= float(int((getEmitterDuration() - delay) / lifetime) + 1) * lifetime));
    v_color.a *= float(shouldDiscard < 0.5);
    v_color.a *= getOvermultiplyAlpha();
    v_color.a *= distanceOpacity;

    v_time.x = realTime;
    v_time.y = t;
}
