

#if !defined(PROJECTION_UTILITY_SHADER)
#define PROJECTION_UTILITY_SHADER

////////////////////////////////////////////////////////////////////////////////////////////////////
    //      P R O J E C T I O N   U T I L I T I E S      //
////////////////////////////////////////////////////////////////////////////////////////////////////

uniform vec4 u_cc3CameraFrustumDepth;	/**< The depth of the camera frustum (far clip, near clip, -(f+n)/(f-n), -2nf/(f-n)). */

/**
 * Given "depthSample" value from a depth buffer, calculate the eqivalent value world space units.
 *
 * Depth values writen into the depth buffer are not linear in world space. They are biased to allow
 * for more precision near the camera.
 *
 * Often, however, we want to know what this corresponds to in linear world space units (ie, which
 * might be meters or centimeters, or whatever). This function converts a value that has been read
 * from a depth buffer into linear world space units.
 */
float depthSampleToWorldSpaceDepth(float depthSample) {
    float zn = 2.0 * depthSample - 1.0;         // Normalized device coordinates
    return u_cc3CameraFrustumDepth.a / (zn + u_cc3CameraFrustumDepth.b);
}

uniform vec4 u_cc3CameraViewport;

float retinaScaleFactor() {
    #if CC3_PLATFORM_IOS
        return 2.0;
    #else
        return 1.0;
    #endif
}

vec2 reciprocalViewportSize() {
    return 1.0 / (u_cc3CameraViewport.zw * retinaScaleFactor());
}


#endif
